"""
This file is not intended to be imported.

Import the sherpa or xspec module instead::

	import bxa.sherpa as bxa

or, for Xspec::

	import bxa.xspec as bxa

"""

__author__ = """Johannes Buchner"""
__email__ = 'johannes.buchner.acad@gmx.com'
__version__ = '5.1.1'
